#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Defines the format of a SPEXone Level-1A product."""

from __future__ import annotations

__all__ = ["init_l1a"]

from typing import TYPE_CHECKING

import numpy as np

# pylint: disable=no-name-in-module
from netCDF4 import Dataset, Variable

from .tmtc_def import tmtc_dtype

if TYPE_CHECKING:
    import datetime as dt

# - global parameters ------------------------------
ORBIT_DURATION = 5904  # seconds


# - local functions --------------------------------
def attrs_sec_per_day(dset: Variable, ref_date: dt.datetime) -> None:
    """Add CF attributes to a dataset holding 'seconds of day'.

    Parameters
    ----------
    dset : Variable
       NetCDF4 variable containing a timestamp as seconds since reference date
    ref_date : dt.datetime
       Reference date

    Examples
    --------
    Update the attributes of variable 'time':

      ref_date = dt.datetime(2022, 3, 21, tzinfo=dt.timezone.utc)
      dset = sgrp.createVariable('image_time', 'f8', ('number_of_images',),
                                fill_value=-32767)
      dset.long_name = "image time"
      dset.description = "Integration start time in seconds of day."
      attrs_sec_per_day(dset, ref_date)

    In CDL the variable `time` will be defined as::

       double time(number_of_scans) ;
          time:_FillValue = -32767 ;
          time:long_name = "time" ;
          time:units = "seconds since 2022-03-21 00:00:00" ;
          time:description = "Earth view mid-time in seconds of day" ;
          time:year = 2022 ;
          time:month = 3 ;
          time:day = 21 ;
          time:valid_min = 0 ;
          time:valid_max = 86401 ;

    Note that '_FillValue', 'long_name' and 'description' are not set by
    this function.

    """
    dset.units = f"seconds since {ref_date.strftime('%Y-%m-%d %H:%M:%S')}"
    dset.year = f"{ref_date.year}"
    dset.month = f"{ref_date.month}"
    dset.day = f"{ref_date.day}"
    dset.valid_min = 0
    dset.valid_max = 86400 + ORBIT_DURATION


def image_attributes(rootgrp: Dataset, ref_date: dt.datetime) -> None:
    """Define group /image_attributes and its datasets."""
    sgrp = rootgrp.createGroup("/image_attributes")
    dset = sgrp.createVariable("icu_time_sec", "u4", ("number_of_images",))
    dset.long_name = "ICU time stamp (seconds)"
    dset.description = "Science TM parameter ICU_TIME_SEC."
    dset.valid_min = np.uint32(1956528000)  # year 2020
    dset.valid_max = np.uint32(2493072000)  # year 2037
    dset.units = "seconds since 1958-01-01 00:00:00 TAI"
    dset = sgrp.createVariable("icu_time_subsec", "u2", ("number_of_images",))
    dset.long_name = "ICU time stamp (sub-seconds)"
    dset.description = "Science TM parameter ICU_TIME_SUBSEC."
    dset.valid_min = np.uint16(0)
    dset.valid_max = np.uint16(0xFFFF)
    dset.units = "1/65536 s"

    dset = sgrp.createVariable(
        "image_time", "f8", ("number_of_images",), fill_value=-32767
    )
    dset.long_name = "image time"
    dset.description = "Integration start time in seconds of day."
    attrs_sec_per_day(dset, ref_date)
    dset = sgrp.createVariable("image_ID", "i4", ("number_of_images",))
    dset.long_name = "image counter from power-up"
    dset.valid_min = np.int32(0)
    dset.valid_max = np.int32(0x7FFFFFFF)
    dset = sgrp.createVariable("binning_table", "u1", ("number_of_images",))
    dset.long_name = "binning-table ID"
    dset.valid_min = np.uint8(0)
    dset.valid_max = np.uint8(0xFF)
    dset = sgrp.createVariable("digital_offset", "i2", ("number_of_images",))
    dset.long_name = "digital offset"
    dset.units = "1"
    dset = sgrp.createVariable(
        "nr_coadditions", "u2", ("number_of_images",), fill_value=0
    )
    dset.long_name = "number of coadditions"
    dset.valid_min = np.int32(1)
    dset.units = "1"
    dset = sgrp.createVariable(
        "exposure_time", "f8", ("number_of_images",), fill_value=0
    )
    dset.long_name = "exposure time"
    dset.units = "s"


def get_chunksizes(ydim: int, compression: bool) -> tuple[int, int]:
    """Obtain chunksizes for dataset: /science_data/science_data."""
    # I did some extensive testing.
    # - Without compression (chunked vs contiguous):
    #   * Writing to a contiguous dataset is faster (10-20%)
    #   * Reading one image is about as fast for chunked and contiguous storage
    #   * Reading a pixel image is much faster for chunked storage (2-8x)
    # - With compression (always chunked):
    #   * Writing takes 3x as long compared to without compression,
    #     but saves about > 40% on disk storage
    #   * Reading of compressed data is much slower than uncompressed data
    #   * The performance when reading one detector image is acceptable,
    #     however reading one pixel image is really slow (specially full-frame).
    # Therefore, these are the best choices for the variable `chunksizes`.
    return (
        (20, ydim)
        if ydim < 1048576
        else (1, min(512 * 1024, ydim))
        if compression
        else (1, ydim)
    )


def science_data(
    rootgrp: Dataset, compression: bool, chunksizes: tuple[int, int]
) -> None:
    """Define group /science_data and its datasets."""
    sgrp = rootgrp.createGroup("/science_data")
    dset = sgrp.createVariable(
        "detector_images",
        "u2",
        ("number_of_images", "samples_per_image"),
        compression="zlib" if compression else None,
        complevel=1,
        chunksizes=chunksizes,
        fill_value=0xFFFF,
    )
    dset.long_name = "detector pixel values"
    dset.valid_min = np.uint16(0)
    dset.valid_max = np.uint16(0xFFFE)
    dset.units = "counts"
    hk_dtype = rootgrp.createCompoundType(tmtc_dtype(0x350), "science_dtype")
    dset = sgrp.createVariable(
        "detector_telemetry", hk_dtype, dimensions=("number_of_images",)
    )
    dset.long_name = "SPEX science telemetry"
    dset.comment = "A subset of MPS and housekeeping parameters."


def engineering_data(rootgrp: Dataset, ref_date: dt.datetime) -> None:
    """Define group /engineering_data and its datasets."""
    sgrp = rootgrp.createGroup("/engineering_data")
    dset = sgrp.createVariable("HK_tlm_time", "f8", ("hk_packets",), fill_value=-32767)
    dset.long_name = "HK telemetry packet time"
    dset.description = "Packaging time in seconds of day."
    attrs_sec_per_day(dset, ref_date)
    hk_dtype = rootgrp.createCompoundType(tmtc_dtype(0x320), "nomhk_dtype")
    dset = sgrp.createVariable("NomHK_telemetry", hk_dtype, ("hk_packets",))
    dset.long_name = "SPEX nominal-HK telemetry"
    dset.comment = "An extended subset of the housekeeping parameters."
    dset = sgrp.createVariable("temp_detector", "f4", ("hk_packets",))
    dset.long_name = "detector temperature"
    dset.comment = "TS1 DEM Temperature (nominal)."
    dset.valid_min = 260
    dset.valid_max = 300
    dset.units = "K"
    dset = sgrp.createVariable("temp_housing", "f4", ("hk_packets",))
    dset.long_name = "housing temperature"
    dset.comment = "TS2 Housing Temperature (nominal)."
    dset.valid_min = 260
    dset.valid_max = 300
    dset.units = "K"
    dset = sgrp.createVariable("temp_radiator", "f4", ("hk_packets",))
    dset.long_name = "radiator temperature"
    dset.comment = "TS3 Radiator Temperature (nominal)."
    dset.valid_min = 260
    dset.valid_max = 300
    dset.units = "K"
    # hk_dtype = rootgrp.createCompoundType(tmtc_dtype(0x322)), 'demhk_dtype')
    # dset = sgrp.createVariable('DemHK_telemetry', hk_dtype, ('hk_packets',))
    # dset.long_name = "SPEX detector-HK telemetry"
    # dset.comment = "DEM housekeeping parameters."


# - main function ----------------------------------
def init_l1a(
    l1a_flname: str, ref_date: dt.datetime, dims: dict, compression: bool = False
) -> Dataset:
    """Create an empty SPEXone Level-1A product (on-ground or in-flight).

    Parameters
    ----------
    l1a_flname : str
       Name of L1A product
    ref_date :  dt.datetime
       Date of the first detector image (UTC)
    dims :  dict
       Provide length of the Level-1A dimensions. Default values::

          number_of_images : None     # number of image frames
          samples_per_image : None    # depends on binning table
          hk_packets : None           # number of HK tlm-packets (1 Hz)

    compression : bool, default=False
       Use compression on dataset /science_data/detector_images.

    Notes
    -----
    The optional groups '/gse_data' and '/navigation_data' are not created
    by this script.

    Original CDL definition is from F. S. Patt (GSFC), 08-Feb-2019

    """
    # check function parameters
    if not isinstance(dims, dict):
        raise TypeError("dims should be a dictionary")

    # initialize dimensions
    number_img = dims.get("number_of_images")
    img_samples = dims.get("samples_per_image")
    hk_packets = dims.get("hk_packets")

    # create/overwrite netCDF4 product
    try:
        rootgrp = Dataset(l1a_flname, "w")
    except Exception as exc:
        raise Exception(f"Failed to create netCDF4 file {l1a_flname}") from exc

    # - define global dimensions
    _ = rootgrp.createDimension("number_of_images", number_img)
    _ = rootgrp.createDimension("samples_per_image", img_samples)
    _ = rootgrp.createDimension("hk_packets", hk_packets)

    # - define the various HDF54/netCDF4 groups and their datasets
    image_attributes(rootgrp, ref_date)
    chunksizes = get_chunksizes(img_samples, compression)
    science_data(rootgrp, compression, chunksizes)
    engineering_data(rootgrp, ref_date)

    return rootgrp



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2022-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
#
"""Provide function to start the logger for spx1_monitor."""

from __future__ import annotations

__all__ = ["start_logger"]

from importlib.resources import files
from logging.config import dictConfig

import yaml


def start_logger() -> None:
    """Initialize logger for pyspex."""
    yaml_fl = files("pyspex.data").joinpath("logger_setup.yaml")
    if not yaml_fl.is_file():
        raise FileNotFoundError(f"{yaml_fl} not found")

    with yaml_fl.open("r", encoding="ascii") as fid:
        try:
            config_data = yaml.safe_load(fid)
        except yaml.YAMLError as exc:
            raise RuntimeError("failed to read YAML file") from exc

    dictConfig(config_data)



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2022-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Obtain settings to generate a L1A product."""

from __future__ import annotations

__all__ = ["argparse_gen_l1a"]

import argparse
import logging
import sys
from dataclasses import dataclass, field
from pathlib import Path

import yaml

from .lib import pyspex_version

# - global parameters ------------------------------
ARG_INPUT_HELP = """provide one or more input files:
- raw: CCSDS packages generated by the SPEXone ICU -- provide name
       of one file with extension '.H'. The files with science and
       housekeeping data are collected using Unix filename pattern
       matching.
- st3: CCSDS packages with ITOS and spacewire headers -- provide name
       of one file with extension '.ST3'.
- dsb: CCSDS packages with PACE headers -- provide list of filenames
       with extension '.spx'.
"""

ARG_YAML_HELP = """provide settings file in YAML format (in-flight example):

# define output directory, default is current working directory
outdir: .
# define name of output file, will be generated automatically when empty
outfile: ''
# compress the dataset /science_data/detector_images
compression: True
# define file-version as nn, neglected when outfile not empty
file_version: 1
# provide processing version
processing_version: '1.0'
# flag to indicate measurements taken in eclipse or day-side
eclipse: True
# provide list, directory, file-glob or empty
hkt_list: <PATH>/PACE.20220617T011*.HKT.nc
# must be a list, directory or glob. Fails when empty
l0_list:
- <PATH>/SPX022000010.spx
- <PATH>/SPX022000011.spx
- <PATH>/SPX022000012.spx
- <PATH>/SPX022000013.spx

"""

_prog_name_ = Path(sys.argv[0]).name
EPILOG_HELP = f"""Usage:
  Generate L1A from OCAL level-0 data directly from the SPEXone instrument:

    {_prog_name_} <Path>/NomSciCal1_20220123T121801.676167.H

    Note that OCAL science & telemetry data is read from the files:
      <Path>/NomSciCal1_20220123T121801.676167.?
      <Path>/NomSciCal1_20220123T121801.676167.??
      <Path>/NomSciCal1_20220123T121801.676167_hk.?

  Generate L1A from OCAL level-0 data via ITOS from the PACE platform:

    {_prog_name_} <Path>/DIAG_20220124_175458_073.ST3

  Generate L1A from in-flight level-0 data, store product in directory L1A:

    {_prog_name_} --outdir L1A <Path>/SPX*.spx

  Generate L1A from in-flight level-0 data read settings from a YAML file:

    {_prog_name_} --yaml config_l1a_gen.yaml

    An example YAML file:
       outdir: L1A
       outfile: ''
       file_version: 1
       processing_version: 1.0
       compression: False
       eclipse: False
       hkt_list: HKT/PACE.20220617T011*.HKT.nc
       l0_list: <PATH>/SPX0220000??.spx

  Dry-run, be extra verbose without generating data:

    {_prog_name_} --debug <Path>/NomSciCal1_20220123T121801.676167.H

  Read level-0 data and dump CCSDS packet headers in ASCII:

    {_prog_name_} --dump <Path>/NomSciCal1_20220123T121801.676167.H

Return codes:
  2      Failed to parse command-line parameters.
  110    One (or more) SPEXone level-0 files not found.
  115    Failed to generate output directry due to permission error.
  121    Input file not recognized as a SPEXone level-0 product.
  122    Corrupted SPEXone level-0 data.
  131    Failed to generate output file due to netCDF4 library issues.
  132    Incomplete set of navigation data detected
  135    Failed to generate output file due to permission error.

Environment:
   'OCVARROOT'
       The number of leap seconds for the TAI to UTC conversion are determined
       using the file 'tai-utc.dat'. A copy of this file is included in the
       package `pyspex`. The latest version can be obtained from
         `https://maia.usno.navy.mil/ser7/tai-utc.dat`.
       When OCVARROOT is set the path should be '$OCVARROOT/common/tai-utc.dat'.
"""


# - local functions --------------------------------
# pylint: disable=too-many-instance-attributes
@dataclass()
class Config:
    """Initiate class to hold settings for L0->L1a processing."""

    debug: bool = False
    dump: bool = False
    verbose: int = logging.NOTSET
    compression: bool = False
    outdir: Path = Path(".").resolve()
    outfile: str = ""
    file_version: int = 1
    processing_version: str = "1.0"
    eclipse: bool | None = None
    yaml_fl: Path = None
    hkt_list: list[Path] = field(default_factory=list)
    l0_format: str = ""
    l0_list: list[Path] = field(default_factory=list)


def __commandline_settings() -> Config:
    """Parse command-line parameters."""

    class NumericLevel(argparse.Action):
        """Store verbosity level of the logger as a numeric value."""

        def __call__(
            self: NumericLevel,
            parser_local: argparse.ArgumentParser,
            namespace: argparse.Namespace,
            values: str,
            option_string: str | None = None,
        ) -> None:
            numeric_level = getattr(logging, values.upper(), None)
            setattr(namespace, self.dest, numeric_level)

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description="Generate PACE level-1A product from SPEXone level-0 data.",
        epilog=EPILOG_HELP,
    )
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s v" + pyspex_version()
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="be extra verbose, no output files generated",
    )
    parser.add_argument(
        "--dump", action="store_true", help="dump CCSDS packet headers in ASCII"
    )
    parser.add_argument(
        "--verbose",
        nargs="?",
        const="info",
        default=logging.WARNING,
        action=NumericLevel,
        choices=("debug", "info", "warning", "error"),
        help='set verbosity level, default is "warning"',
    )
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument(
        "--eclipse",
        action="store_true",
        default=None,
        help="assume that measurements are perfomed in eclipse",
    )
    group.add_argument(
        "--no_eclipse",
        dest="eclipse",
        action="store_false",
        help="assume that measurements are not perfomed in eclipse",
    )
    parser.add_argument(
        "--outdir",
        type=Path,
        default=None,
        help="directory to store the generated level-1A product(s)",
    )
    # group = parser.add_mutually_exclusive_group(required=True)
    parser.add_argument("--yaml", type=Path, default=None, help=ARG_YAML_HELP)
    parser.add_argument("lv0_list", nargs="*", help=ARG_INPUT_HELP)
    args = parser.parse_args()

    config = Config()
    if args.debug:
        config.debug = True
    if args.dump:
        config.dump = True
    if args.verbose:
        config.verbose = args.verbose
    if args.eclipse is not None:
        config.eclipse = args.eclipse
    if args.outdir is not None:
        config.outdir = args.outdir
    if args.yaml:
        config.yaml_fl = args.yaml
    elif args.lv0_list:
        config.l0_list = [Path(x) for x in args.lv0_list]
    else:
        parser.error("You should provide a YAML file or names of L0 products")

    return config


# pylint: disable=too-many-branches
def __yaml_settings(config: dataclass) -> dataclass:
    """Read YAML configuration file."""
    with open(config.yaml_fl, encoding="ascii") as fid:
        config_yaml = yaml.safe_load(fid)

    if "outdir" in config_yaml and config_yaml["outdir"] is not None:
        config.outdir = Path(config_yaml["outdir"])
    if "outfile" in config_yaml and config_yaml["outfile"]:
        config.outfile = config_yaml["outfile"]
    if "compression" in config_yaml and config_yaml["compression"]:
        config.compression = True
    if "file_version" in config_yaml and config_yaml["file_version"] != 1:
        config.file_version = config_yaml["file_version"]
    if "processing_version" in config_yaml and config_yaml["processing_version"] != "":
        config.processing_version = config_yaml["processing_version"]
    if "eclipse" in config_yaml and config_yaml["eclipse"] is not None:
        config.eclipse = config_yaml["eclipse"]
    if "hkt_list" in config_yaml and config_yaml["hkt_list"]:
        if isinstance(config_yaml["hkt_list"], list):
            config.hkt_list = [Path(x) for x in config_yaml["hkt_list"]]
        else:
            mypath = Path(config_yaml["hkt_list"])
            if mypath.is_dir():
                config.hkt_list = sorted(Path(mypath).glob("*"))
            else:
                config.hkt_list = sorted(Path(mypath.parent).glob(mypath.name))
    if "l0_list" in config_yaml and config_yaml["l0_list"]:
        if isinstance(config_yaml["l0_list"], list):
            config.l0_list = [Path(x) for x in config_yaml["l0_list"]]
        else:
            mypath = Path(config_yaml["l0_list"])
            if mypath.is_dir():
                config.l0_list = sorted(Path(mypath).glob("*"))
            else:
                config.l0_list = sorted(Path(mypath.parent).glob(mypath.name))

    return config


# - main function ----------------------------------
def argparse_gen_l1a() -> dataclass:
    """Obtain settings from both command-line and YAML file (when provided).

    Returns
    -------
    dataclass
       settings from both command-line arguments and YAML config-file

    """
    config = __commandline_settings()
    if config.yaml_fl is None:
        return config

    if not config.yaml_fl.is_file():
        raise FileNotFoundError(config.yaml_fl)

    return __yaml_settings(config)



#!/usr/bin/env python3
#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2022-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Command-line implementation of spx1_level01a."""

from __future__ import annotations

import logging
import warnings

from pyspex.argparse_gen_l1a import argparse_gen_l1a
from pyspex.lib.check_input_files import check_input_files
from pyspex.lib.logger import start_logger
from pyspex.lv0_lib import CorruptPacketWarning
from pyspex.tlm import SPXtlm


def main() -> int:
    """Execute the main bit of the application."""
    error_code = 0
    warn_code = 0

    # initialize logger
    start_logger()
    logging.captureWarnings(True)

    # parse command-line parameters and YAML file for settings
    config = argparse_gen_l1a()
    logging.getLogger().setLevel(config.verbose)  # first, set the root logger
    logger = logging.getLogger("pyspex.gen_l1a")  # then initiate a descendant
    logger.debug("%s", config)

    # check input files (SEPXone level-0)
    try:
        check_input_files(config)
    except FileNotFoundError as exc:
        logger.fatal('File "%s" not found on system.', exc)
        return 110
    except TypeError as exc:
        logger.fatal("%s", exc)
        return 121

    # read level 0 data
    # Note that we read as much as possible packages from a file, but stop at
    # the first occurence of a corrupted data-packet, then the remainder of
    # the file is neglected.
    tlm = None
    with warnings.catch_warnings(record=True) as wrec_list:
        warnings.simplefilter("always", category=CorruptPacketWarning)
        try:
            tlm = SPXtlm()
            tlm.from_lv0(
                config.l0_list,
                file_format=config.l0_format,
                debug=config.debug,
                dump=config.dump,
            )
        except FileNotFoundError as exc:
            logger.fatal('FileNotFoundError exception raised for "%s".', exc)
            error_code = 110
        except TypeError as exc:
            logger.fatal('TypeError exception raised with "%s".', exc)
            error_code = 121

        for wrec in wrec_list:
            logger.warning('CorruptPacketWarning raised with "%s".', str(wrec.message))
            warn_code = 122

    if error_code != 0 or config.debug or config.dump:
        return error_code

    # Write Level-1A product.
    if not config.outdir.is_dir():
        config.outdir.mkdir(mode=0o755, parents=True)
    try:
        if config.eclipse is None or tlm.science.size == 0:
            tlm.gen_l1a(config)
        elif config.eclipse:
            tlm.full().gen_l1a(config)
            tlm.binned().gen_l1a(config)
        else:
            tlm.binned().gen_l1a(config)
    except (KeyError, RuntimeError) as exc:
        # raise RuntimeError from exc
        logger.fatal('RuntimeError with "%s"', exc)
        error_code = 131
    except UserWarning as exc:
        logger.warning('navigation data is incomplete: "%s".', exc)
        error_code = 132
    except Exception as exc:
        # raise RuntimeError from exc
        logger.fatal('Unexpected exception occurred with "%s".', exc)
        error_code = 135

    return warn_code if error_code == 0 else error_code


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Add ITOS EGSE information of OCAL measurements to one SPEXone L1A product."""

from __future__ import annotations

import argparse
from pathlib import Path

from pyspex.egse_db import add_egse_data, create_egse_db

# - global parameters ------------------------------


# - local functions --------------------------------


# - main function ----------------------------------
def main() -> None:
    """Add EGSE data to an existing SPEXone L1A product."""
    # parse command-line parameters
    parser = argparse.ArgumentParser()
    parser.add_argument("--verbose", action="store_true", help="be verbose")
    parser.add_argument(
        "--egse_dir", default="Logs", type=Path, help="directory with EGSE data"
    )
    subparsers = parser.add_subparsers(help="sub-command help")
    parser_db = subparsers.add_parser("create_db", help="create new EGSE database")
    parser_db.add_argument(
        "file_list", nargs="+", help="provide names EGSE files (CSV)"
    )
    parser_db.set_defaults(func=create_egse_db)

    parser_wr = subparsers.add_parser(
        "add", help=("add EGSE information to a" " SPEXone level-1A product")
    )
    parser_wr.add_argument(
        "l1a_file", default=None, type=str, help="SPEXone L1A product"
    )
    parser_wr.set_defaults(func=add_egse_data)
    args = parser.parse_args()
    if args.verbose:
        print(args)

    # call whatever function was selected
    args.func(args)


# --------------------------------------------------
if __name__ == "__main__":
    main()



# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#   All Rights Reserved
#
# License:  BSD-3-Clause
"""SRON Python package `pyspex`.

It contains software to read PACE HKT products and SPEXone Level-0 products,
and read/write SPEXone Level-1A products.
"""



#!/usr/bin/env python3
#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Convert SPEXone DEM measurements to L1A format.

The DEM measurements are performed at SRON between 17-10-2019 and 13-12-2019.

References
----------
* SRON-SPEX-TN-2020-001_0_5_SPEXone_Detector_Characterization.pdf


"""

from __future__ import annotations

import argparse
from datetime import UTC, datetime, timedelta
from pathlib import Path

import h5py
import numpy as np
from pyspex.dem_io import DEMio, img_sec_of_day
from pyspex.l1a_io import L1Aio
from pyspex.lib import pyspex_version
from pyspex.lib.tmtc_def import tmtc_dtype
from pyspex.lv1_gse import LV1gse

# - global parameters ------------------------------
DEFAULT_EPOCH = datetime(1970, 1, 1, tzinfo=UTC)


# - local functions --------------------------------
def get_l1a_name(msm_id: str, utc_sensing_start: datetime) -> str:
    """Return name of SPEXone product for DEM measurements.

    Parameters
    ----------
    msm_id : string, optional
       Provide identifier for measurement, OCAL only
    utc_sensing_start: datetime
       Provide sensing start of first measurement in L1A product

    Notes
    -----
    L1A file name format:
       SPX1_OCAL_<msm_id>[_YYYYMMDDTHHMMSS]_L1A_vvvvvvv.nc
    where
       msm_id is the measurement identifier
       YYYYMMDDTHHMMSS is time stamp of the first image in the file
       vvvvvvv is the git-hash string of the pyspex repository

    """
    # define string of sensing start as yyyymmddThhmmss
    sensing_start = utc_sensing_start.strftime("%Y%m%dT%H%M%S")

    return (
        f"SPX1_OCAL_{msm_id}_{sensing_start}" f"_L1A_{pyspex_version(githash=True)}.nc"
    )


# - main function ----------------------------------
def main() -> None:
    """Create a SPEXone L1A product from data of a DEM measurement."""
    parser = argparse.ArgumentParser(
        description="create SPEXone L1A product from DEM measurement(s)"
    )
    parser.add_argument(
        "--verbose",
        default=False,
        action="store_true",
        help="be verbose, default be silent",
    )
    parser.add_argument(
        "--reference", default=None, help="reference detector data (for non-linearity)"
    )
    parser.add_argument(
        "--output", default=None, help="define output directory, default=CWD"
    )
    parser.add_argument(
        "--dem_id",
        choices=("D35", "D39", "D84"),
        help=("provide DEM ID" " or ID will be extracted from path"),
    )
    parser.add_argument(
        "file_list",
        nargs="+",
        help=(
            "provide path to _b.bin file of one or more"
            " SPEXone DEM characterization measurements"
        ),
    )
    args = parser.parse_args()
    if args.verbose:
        print(args)

    # sort file on timestamp in filename (problem: midnight...)
    buff = [name for name in args.file_list if name.endswith(".bin")]
    args.file_list = sorted(buff, key=lambda name: name.split("_")[-2])

    # obtain DEM_ID (can be specified on command-line)
    # and list of measurements used to generate the L1A product
    if args.dem_id is None:
        dem_id_list = []
        for file in args.file_list:
            parts = Path(file).parts
            if len(parts) == 1:
                continue
            for dem_id in ("D35", "D39", "D84"):
                if parts[-2].find(dem_id) != -1:
                    dem_id_list.append(dem_id)

        if not dem_id_list:
            raise KeyError("Can not determine DEM_ID, please specify --dem_id")
        dem_id_list = set(dem_id_list)
        if len(dem_id_list) != 1:
            raise KeyError("DEM_ID of measurements must be unique")

        args.dem_id = dem_id_list.pop()

    # initialze data arrays
    msm_id = None
    tstamp = []
    image_list = []

    # Measurement Parameters Settings
    n_images = len(args.file_list)
    img_hk = np.zeros(n_images, dtype=tmtc_dtype(0x350))
    t_exp = np.empty(n_images, dtype=float)
    t_frm = np.empty(n_images, dtype=float)
    offset = np.empty(n_images, dtype=float)

    for ii, flname in enumerate(args.file_list):
        dem_file = Path(flname)
        if not dem_file.is_file():
            raise FileNotFoundError(f"file {flname} does not exist")
        if dem_file.suffix == ".txt":
            raise ValueError("We should not try a file with suffix '.txt'")

        # read DEM settings and data
        parts = dem_file.name.split("_")
        if len(parts) == 8:
            _id = "_".join(parts[:4])
            tstamp.append(
                datetime.strptime(parts[5] + parts[6] + "+00:00", "%Y%m%d%H%M%S.%f%z")
            )
        elif len(parts) == 10:
            _id = "_".join(parts[:6])
            tstamp.append(
                datetime.strptime(parts[7] + parts[8] + "+00:00", "%Y%m%d%H%M%S.%f%z")
            )
        elif len(parts) == 7:
            _id = "_".join(parts[:3])
            tstamp.append(
                datetime.strptime(parts[4] + parts[5] + "+00:00", "%Y%m%d%H%M%S.%f%z")
            )
        elif len(parts) == 9:
            _id = "_".join(parts[:6])
            tstamp.append(
                datetime.strptime(parts[6] + parts[7] + "+00:00", "%Y%m%d%H%M%S.%f%z")
            )
        else:
            raise ValueError("Invalid format of data-product name")

        if msm_id is None:
            msm_id = _id
        elif msm_id != _id:
            raise ValueError("Do not combine different kind of measurements")

        dem = DEMio(flname)
        # obtain Science_HK information from header file (ASCII)
        img_hk[ii] = dem.get_sci_hk()
        # get nr_coaddings from file name
        coad_str = [x for x in parts if x.startswith("coad")][0]
        img_hk[ii]["REG_NCOADDFRAMES"] = int(coad_str[-2:])
        # determine exposure time
        t_exp[ii] = dem.exp_time()
        t_frm[ii] = dem.frame_period(int(coad_str[-2:]))
        offset[ii] = dem.offset
        # read image data
        image_list.append(dem.get_data())

    images = np.array(image_list)
    del image_list

    # sort data according to timestamps
    if tstamp != sorted(tstamp):
        print("[WARNING]: do a sort of all the data...")
        indx = sorted(range(n_images), key=tstamp.__getitem__)
        img_hk = img_hk[indx]
        tstamp = [tstamp[k] for k in indx]
        images = images[indx]

    # convert timestamps to seconds per day
    img_sec = np.empty(n_images, dtype="u4")
    img_subsec = np.empty(n_images, dtype="u2")
    for ii, tval in enumerate(tstamp):
        img_sec[ii] = (tval - DEFAULT_EPOCH).total_seconds()
        img_subsec[ii] = tval.microsecond * 2**16 // 1000000

    ref_date, img_time = img_sec_of_day(img_sec, img_subsec, img_hk)

    # generate name of L1A product
    prod_name = get_l1a_name(msm_id.strip(" "), tstamp[0])
    if args.output is not None:
        dest_dir = Path(args.output)
        if not dest_dir.is_dir():
            dest_dir.mkdir()
        prod_name = str(dest_dir / prod_name)

    # set dimensions of L1A datasets
    n_samples = images.size if images.ndim == 2 else images.shape[1] * images.shape[2]

    dims = {
        "number_of_images": n_images,
        "samples_per_image": n_samples,
        "hk_packets": 0,
    }

    # generate L1A product
    with L1Aio(prod_name, dims=dims, ref_date=ref_date) as l1a:
        # write image data, detector telemetry and image attributes
        l1a.fill_science(
            images.reshape(n_images, n_samples), img_hk, np.arange(n_images)
        )

        l1a.set_dset("/image_attributes/icu_time_sec", img_sec)
        l1a.set_attr(
            "valid_min", np.uint32(1577800000), ds_name="/image_attributes/icu_time_sec"
        )
        l1a.set_attr(
            "valid_max", np.uint32(1735700000), ds_name="/image_attributes/icu_time_sec"
        )
        l1a.set_attr(
            "units",
            "seconds since 1970-01-01 00:00:00",
            ds_name="/image_attributes/icu_time_sec",
        )
        l1a.set_dset("/image_attributes/icu_time_subsec", img_subsec)
        l1a.set_dset("/image_attributes/image_time", img_time)

        # Global attributes
        l1a.fill_global_attrs(inflight=False)
        if len(tstamp) == 1:
            tstamp.append(tstamp[0] + timedelta(seconds=t_frm[0]))
        l1a.set_attr(
            "time_coverage_start", tstamp[0].isoformat(timespec="milliseconds")
        )
        l1a.set_attr("time_coverage_end", tstamp[-1].isoformat(timespec="milliseconds"))
        l1a.set_attr("input_files", [Path(x).name for x in args.file_list])

    # Add OGSE and EGSE parameters
    # - Light source
    # - DoLP
    # - AoLP
    # - FOV_begin
    # - FOV_end
    # - Detector illumination
    # - Illumination_level
    # - viewport(s)
    # - Wavelength and signal of data stimulus
    with LV1gse(prod_name) as gse:
        gse.set_attr("origin", "SPEXone Detector Characterization (SRON)")
        gse.set_attr("measurement", msm_id)
        gse.set_attr("dem_id", args.dem_id)
        if args.reference is not None:
            sec_bgn = int(tstamp[0].timestamp())
            sec_end = int(tstamp[-1].timestamp())
            with h5py.File(args.reference, "r") as fid:
                secnd = fid["sec"][:]
                data = fid["amps"][((secnd >= sec_bgn) & (secnd <= sec_end))]
            gse.write_reference_signal(np.median(data), data.std())


# --------------------------------------------------
if __name__ == "__main__":
    main()



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Collection of routines to access EGSE data generated by ITOS."""

from __future__ import annotations

__all__ = ["create_egse_db", "add_egse_data"]

from datetime import UTC, datetime, timedelta
from pathlib import Path
from typing import TYPE_CHECKING

import h5py
import numpy as np

# pylint: disable=no-name-in-module
from netCDF4 import Dataset

if TYPE_CHECKING:
    import argparse

# - global parameters ------------------------------
DB_EGSE = "egse_db_itos.nc"

# enumerate source status
LDLS_DICT = {
    b"UNPLUGGED": 0,
    b"Controller Fault": 1,
    b"Idle": 2,
    b"Laser ON": 3,
    b"Lamp ON": 4,
    b"MISSING": 255,
}

# enumerate shutter positions
SHUTTER_DICT = {b"CLOSED": 0, b"OPEN": 1, b"PARTIAL": 255}


# - local functions --------------------------------
def init_gse_data(fid: h5py.File) -> h5py.Group:
    """Initialize the netCDF4 group 'gse_data' for EGSE/OGSE information."""
    # investigate filename
    parts = fid.product_name.split("_")
    if len(parts) > 2 and parts[0] == "SPX1":
        parts = parts[2:]
    msmt_fields = parts[0].split("-")
    background = "BKG" in msmt_fields
    act_angle = [float(x.replace("act", "")) for x in parts if x.startswith("act")]
    alt_angle = [float(x.replace("alt", "")) for x in parts if x.startswith("alt")]
    pol_angle = [
        float(x.replace("pol", ""))
        for x in parts
        if x.startswith("pol") and x != "polcal"
    ]
    gp1_angle = [float(x.replace("glass", "")) for x in parts if x.startswith("glass")]
    gp1_offs = 5.634375
    # gp2_offs = 5.09625

    # determine viewport: default 0, when all viewports are illuminated
    if alt_angle:
        vp_angle = np.array([-50.0, -20, 0, 20, 50])
        vp_diff = np.abs(vp_angle - alt_angle[0])
        viewport = 0 if vp_diff.min() > 6 else 2 ** np.argmin(vp_diff)
    else:
        viewport = 0

    gid = fid.createGroup("/gse_data")
    dset = gid.createVariable("viewport", "u1")
    dset.long_name = "viewport status"
    dset.standard_name = "status_flag"
    dset.valid_range = np.array([0, 16], dtype="u1")
    dset.flag_values = np.array([0, 1, 2, 4, 8, 16], dtype="u1")
    dset.flag_meanings = "ALL -50deg -20deg 0deg +20deg +50deg"
    dset[:] = viewport

    # gid.FOV_begin = np.nan
    # gid.FOV_end = np.nan
    gid.ACT_rotationAngle = np.nan if not act_angle else act_angle[0]
    gid.ALT_rotationAngle = np.nan if not alt_angle else alt_angle[0]
    # gid.ACT_illumination = np.nan
    # gid.ALT_illumination = np.nan
    gid.AoLP = 0.0 if not pol_angle else pol_angle[0]
    if not background and msmt_fields[0] in ("POLARIZED", "POLARIMETRIC"):
        gid.DoLP = 1.0
        if gp1_angle:
            gid.GP1_angle = gp1_angle[0] + gp1_offs
    else:
        gid.DoLP = 0.0

    return gid


def byte_to_timestamp(str_date: bytes) -> float:
    """Convert a byte-string to a timestamp."""
    buff = str_date.decode("ascii").strip() + "+00:00"  # date is in UTC
    return datetime.strptime(buff, "%Y%m%dT%H%M%S.%f%z").timestamp()


def egse_dtype() -> np.dtype:
    """Define numpy structured array to hold EGSE data."""
    return np.dtype(
        [
            ("ITOS_time", "f8"),
            ("NOMHK_packets_time", "f8"),
            ("LDLS_STATUS", "u1"),
            ("POLARIZER_MOVING", "u1"),
            ("SHUTTER_STAGE_MOVING", "u1"),
            ("STST_POLARIZER_MOVING", "u1"),
            ("ALT_STAGE_MOVING", "u1"),
            ("ACT_STAGE_MOVING", "u1"),
            ("GP_0_MOVING", "u1"),
            ("GP_1_MOVING", "u1"),
            ("C_FLEX-405nm", "u1"),
            ("C_FLEX-457nm", "u1"),
            ("C_FLEX-515nm", "u1"),
            ("C_FLEX-561nm", "u1"),
            ("C_FLEX-660nm", "u1"),
            ("COBOLT-785nm", "u1"),
            ("CRYSTA_STATUS", "u1"),  # should be C_FLEX-732nm!
            # 15 bytes, will be aligned at 16 or 18 bytes
            ("HK_TS1", "u4"),
            ("HK_TS2", "u4"),
            ("SURV_TST1", "f4"),
            ("SURV_TST2", "f4"),
            ("AI_02", "f4"),
            ("AI_03", "f4"),
            ("AI_04", "f4"),
            ("AI_05", "f4"),
            ("AI_06", "f4"),
            ("AI_07", "f4"),
            ("AI_08", "f4"),
            ("AI_09", "f4"),
            ("V_OUT_ICU", "f4"),
            ("I_OUT_ICU", "f4"),
            ("V_OUT_HTR", "f4"),
            ("I_OUT_HTR", "f4"),
            ("POLARIZER", "f4"),
            ("SHUTTER_STAGE", "f4"),
            ("STST_POLARIZER", "f4"),
            ("ALT_ANGLE", "f4"),
            ("ACT_ANGLE", "f4"),
            ("GP_0_ANGLE", "f4"),
            ("GP_1_ANGLE", "f4"),
        ]
    )


def egse_units() -> tuple[str, ...]:
    """Define numpy structured array to hold EGSE data."""
    return (
        "s",
        "s",
        "1",
        "1",
        "1",
        "1",
        "1",
        "1",
        "1",
        "1",
        "1",
        "1",
        "1",
        "1",
        "1",
        "1",
        "1",
        "Ohm",
        "Ohm",
        "V",
        "V",
        "V",
        "V",
        "V",
        "V",
        "V",
        "V",
        "V",
        "V",
        "V",
        "A",
        "V",
        "A",
        "deg",
        "deg",
        "deg",
        "deg",
        "deg",
        "deg",
        "deg",
    )


def read_egse(
    egse_file: str, verbose: bool = False
) -> tuple[np.ndarray, list[str, ...]]:
    """Read EGSE data (tab separated values) to numpy compound array."""
    with open(egse_file, encoding="ascii") as fid:
        line = None
        names = []
        units = []
        while not line:
            line = fid.readline().strip()
            fields = line.split("\t")
            for field in fields:
                if field == "":
                    continue
                res = field.strip().split(" [")
                names.append(res[0].replace(" nm", "nm").replace(" ", "_"))
                if len(res) == 2:
                    units.append(res[1].replace("[", "").replace("]", ""))
                else:
                    units.append("1")

        if len(names) in (35, 36):
            # define dtype of the data
            formats = (
                ("f8",)
                + 14 * ("f4",)
                + ("u1",)
                + 2 * ("i4",)
                + (
                    "f4",
                    "u1",
                )
                + 2 * ("u1",)
                + 3
                * (
                    "f4",
                    "u1",
                )
                + 7 * ("u1",)
            )
        else:
            # define dtype of the data
            formats = (
                ("f8",)
                + 14 * ("f4",)
                + ("u1",)
                + 2 * ("i4",)
                + (
                    "f4",
                    "u1",
                )
                + 2 * ("u1",)
                + 5
                * (
                    "f4",
                    "u1",
                )
                + 7 * ("u1",)
            )
        usecols = list(range(len(names)))

        if "NOMHK_packets_time" in names:
            formats = ("f8",) + formats
            convertors = {
                0: byte_to_timestamp,
                1: byte_to_timestamp,
                16: lambda s: LDLS_DICT.get(s.strip(), 255),
                21: lambda s: SHUTTER_DICT.get(s.strip(), 255),
            }
        else:
            convertors = {
                0: byte_to_timestamp,
                15: lambda s: LDLS_DICT.get(s.strip(), 255),
                20: lambda s: SHUTTER_DICT.get(s.strip(), 255),
            }
        if verbose:
            print(len(names), names)
            print(len(units), units)
            print(len(formats), formats)

        if not len(names) == len(units) == len(formats):
            raise RuntimeError("Size of names, units or formats are not equal")

        data = np.loadtxt(
            fid,
            delimiter="\t",
            converters=convertors,
            usecols=usecols,
            dtype={"names": names, "formats": formats},
        )

    egse = np.empty(data.size, dtype=egse_dtype())
    egse["NOMHK_packets_time"][:] = np.nan
    egse["GP_0_ANGLE"][:] = np.nan
    egse["GP_1_ANGLE"][:] = np.nan
    egse["GP_0_MOVING"][:] = 255
    egse["GP_1_MOVING"][:] = 255
    for name in data.dtype.names:
        egse[name][:] = data[name][:]

    return egse, units


# ---------- CREATE EGSE DATABASE ----------
def create_egse_db(args: argparse.Namespace) -> None:
    """Write EGSE data to HDF5 database."""
    egse = None
    for egse_file in args.file_list:
        try:
            res = read_egse(egse_file, verbose=args.verbose)
        except RuntimeError:
            return

        egse = res[0] if egse is None else np.concatenate((egse, res[0]))

    with Dataset(args.egse_dir / DB_EGSE, "w", format="NETCDF4") as fid:
        fid.input_files = [Path(x).name for x in args.file_list]
        fid.creation_date = datetime.now(UTC).isoformat(timespec="seconds")

        _ = fid.createEnumType(
            "u1",
            "ldls_t",
            {k.replace(b" ", b"_").upper(): v for k, v in LDLS_DICT.items()},
        )
        _ = fid.createEnumType(
            "u1", "shutter_t", {k.upper(): v for k, v in SHUTTER_DICT.items()}
        )
        _ = fid.createDimension("time", egse.size)
        dset = fid.createVariable("time", "f8", ("time",), chunksizes=(256,))
        time_key = "ITOS_time" if "ITOS_time" in egse.dtype.names else "time"
        indx = np.argsort(egse[time_key])
        dset[:] = egse[time_key][indx]

        egse_t = fid.createCompoundType(egse.dtype, "egse_dtype")
        dset = fid.createVariable("egse", egse_t, ("time",), chunksizes=(64,))
        dset.long_name = "EGSE settings"
        dset.fields = np.bytes_(egse.dtype.names)
        dset.units = np.bytes_(egse_units())
        dset.comment = (
            "DIG_IN_00 is of enumType ldls_t;"
            " SHUTTER_STATUS is of enumType shutter_t"
        )
        dset[:] = egse[indx]


# ----- SELECT OGSE DATA FROM DATABASE AND ADD TO L1A PRODUCT -----
def add_egse_data(args: argparse.Namespace) -> None:
    """Write EGSE records of a measurement to a level-1A product."""
    # determine duration of the measurement (ITOS clock)
    with h5py.File(args.l1a_file, "r") as fid:
        # pylint: disable=unsubscriptable-object
        res = fid.attrs["input_files"]
        if isinstance(res, bytes):
            input_file = Path(res.decode("ascii")).stem.rstrip("_hk")
        else:
            input_file = Path(res[0]).stem.rstrip("_hk")
        # pylint: disable=no-member
        msmt_start = datetime.fromisoformat(
            fid.attrs["time_coverage_start"].decode("ascii")
        )
        msmt_stop = datetime.fromisoformat(
            fid.attrs["time_coverage_end"].decode("ascii")
        )
        if args.verbose:
            print(
                "L1A time-coverage:",
                fid.attrs["time_coverage_start"].decode("ascii"),
                fid.attrs["time_coverage_end"].decode("ascii"),
            )
        duration = np.ceil((msmt_stop - msmt_start).total_seconds())

    # use the timestamp in the filename to correct ICU time
    date_str = input_file.split("_")[-1] + "+00:00"
    msmt_start = datetime.strptime(date_str, "%Y%m%dT%H%M%S.%f%z")
    msmt_start = msmt_start.replace(microsecond=0)
    msmt_stop = msmt_start + timedelta(seconds=int(duration))
    if args.verbose:
        print("Corrected time-coverage:", msmt_start.timestamp(), msmt_stop.timestamp())

    # open EGSE database
    with Dataset(args.egse_dir / DB_EGSE, "r") as fid:
        egse_time = fid["time"][:].data
        if args.verbose:
            print("EGSE time-coverage:", egse_time.min(), egse_time.max())
        mask = (egse_time >= msmt_start.timestamp()) & (
            egse_time <= msmt_stop.timestamp()
        )
        if mask.sum() == 0:
            raise RuntimeError("no EGSE data found")

        egse_time = egse_time[mask]
        egse_data = fid["egse"][mask]

    # update Level-1A product with EGSE information
    with Dataset(args.l1a_file, "r+") as fid:
        gid = fid["/gse_data"] if fid.groups.get("/gse_data") else init_gse_data(fid)
        _ = gid.createEnumType(
            "u1",
            "ldls_t",
            {k.replace(b" ", b"_").upper(): v for k, v in LDLS_DICT.items()},
        )
        _ = gid.createEnumType(
            "u1", "shutter_t", {k.upper(): v for k, v in SHUTTER_DICT.items()}
        )
        _ = gid.createDimension("time", egse_data.size)
        dset = gid.createVariable("time", "f8", ("time",))
        dset[:] = egse_time

        egse_t = gid.createCompoundType(egse_data.dtype, "egse_dtype")
        dset = gid.createVariable("egse", egse_t, ("time",))
        dset.long_name = "EGSE settings"
        dset.fields = np.bytes_(egse_data.dtype.names)
        dset.units = np.bytes_(egse_units())
        dset.comment = (
            "DIG_IN_00 is of enumType ldls_t;"
            " SHUTTER_STATUS is of enumType shutter_t"
        )
        dset[:] = egse_data

        def smart_average(data: np.ndarray, thres_range: float = 0.1) -> float:
            val_range = np.abs(data.max() - data.min())
            if val_range == 0:
                return float(data[0])

            if val_range < thres_range:
                return np.mean(data)

            return np.median(data)

        fid["/gse_data"].ACT_rotationAngle = smart_average(egse_data["ACT_ANGLE"])
        fid["/gse_data"].ALT_rotationAngle = smart_average(egse_data["ALT_ANGLE"])
        fid["/gse_data"].AoLP = smart_average(egse_data["POLARIZER"])



#!/usr/bin/env python3
#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Quick and dirty script to generate simple Quick-Look figures."""

from __future__ import annotations

import argparse
from datetime import datetime, timedelta
from pathlib import Path

import h5py
import numpy as np
from moniplot.draw_image import DrawImage
from moniplot.lib.fig_info import FIGinfo
from moniplot.mon_plot import MONplot
from pyspex.binning_tables import BinningTables


# --------------------------------------------------
def binned_to_2x2_image(table_id: int, img_binned: np.ndarray) -> np.ndarray:
    """Convert binned detector data to image (1024, 1024)."""
    try:
        bin_ckd = BinningTables()
    except Exception as exc:
        raise RuntimeError from exc

    return bin_ckd.unbin(table_id, img_binned).reshape(1024, 1024)


# --------------------------------------------------
def main() -> None:
    """Create a PDF from a SPEXone Level-1A product."""
    parser = argparse.ArgumentParser(
        description="create Quick-Look from SPEXone L1A product"
    )
    parser.add_argument(
        "--verbose",
        default=False,
        action="store_true",
        help="be verbose, default be silent",
    )
    parser.add_argument(
        "--show_images",
        type=str,
        default=None,
        help="comma separated list, default use --max_images",
    )
    parser.add_argument(
        "--max_images",
        type=int,
        default=20,
        help="maximum number of images in quick-look (20)",
    )
    parser.add_argument("file_list", nargs="*", help="provide name of L1A product")
    args = parser.parse_args()
    if args.verbose:
        print(args)

    if args.file_list:
        file_list = [Path(xx) for xx in args.file_list]
    else:
        file_list = Path(".").glob("*.nc")

    for flname in sorted(file_list):
        print(flname)

        with h5py.File(flname, "r") as fid:
            coverage_start = fid.attrs["time_coverage_start"]
            image_time = fid["/image_attributes/image_time"][:]
            exposure_time = fid["/image_attributes/exposure_time"][:]
            table_id = fid["/image_attributes/binning_table"][:]
            sci_hk = fid["/science_data/detector_telemetry"][:]
            dset = fid["/science_data/detector_images"]
            # pylint: disable=no-member
            images = dset.astype(float)[:]
        try:
            coverage_start = coverage_start.decode()
        except (UnicodeDecodeError, AttributeError):
            date_start = "1970-01-01T00:00:00"
        else:
            date_start = coverage_start.split(".")[0]

        data_dir = flname.parent / "QuickLook"
        if not data_dir.is_dir():
            data_dir.mkdir(mode=0o755)

        # open plot object
        report = MONplot((data_dir / flname.name).with_suffix(".pdf"))
        report.set_institute("SRON")

        # which images are requested?
        n_img = images.shape[0]
        if args.show_images is None:
            if n_img <= args.max_images:
                indx = range(n_img)
            else:
                indx = np.arange(0, n_img, n_img / args.max_images)
                indx = [int(np.ceil(x)) for x in indx]
        else:
            if args.show_images == "all":
                indx = range(n_img)
            else:
                indx = [int(x) for x in args.show_images.split(",")]

        med_table_id = np.sort(table_id)[len(table_id) // 2]

        # generate pages in quick-look
        for ii in indx:
            if med_table_id > 0:
                img2d = binned_to_2x2_image(med_table_id, images[ii, :]) / 4
            else:
                if images[ii, :].size != 4194304:
                    continue
                img2d = images[ii, :].reshape(2048, 2048)
            img2d /= sci_hk[ii]["REG_NCOADDFRAMES"]

            time_str = (
                datetime(year=2020, month=1, day=1) + timedelta(seconds=image_time[ii])
            ).strftime("%H:%M:%S.%f")

            figinfo = FIGinfo()
            figinfo.add("coverage_start", date_start)
            figinfo.add("image_time", time_str)
            figinfo.add("exposure_time", exposure_time[ii], fmt="{:f}s")
            figinfo.add(
                "signal_range",
                (np.nanmin(img2d), np.nanmax(img2d)),
                fmt="[{:.2f}, {:.2f}]",
            )
            if table_id[ii] > 0:
                suptitle = f"frame [table_id={table_id[ii]}]: {ii}"
            else:
                suptitle = f"frame: {ii}"
            # plot.set_cmap(tol_cmap("rainbow_WhBr"))
            plot = DrawImage(img2d)
            fig, axx = plot.subplots()
            plot.draw(axx, fig_info=figinfo, title=suptitle)
            report.add_copyright(axx["image"])
            report.close_this_page(fig, None)

        # close plot object
        report.close()


# --------------------------------------------------
if __name__ == "__main__":
    main()



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Contains the class `CCSDSio` to read SPEXone telemetry packets."""

from __future__ import annotations

__all__ = ["CCSDSio", "img_sec_of_day", "hk_sec_of_day"]

from datetime import UTC, datetime, timedelta
from pathlib import Path
from typing import Any

import numpy as np

from .lib.tmtc_def import tmtc_dtype

# - global parameters ------------------------------
FULLFRAME_BYTES = 2 * 2048 * 2048

# Define parameters of Primary header
#  - Packet type     (3 bits): Version No.
#                              Indicates this is a CCSDS version 1 packet
#                     (1 bit): Type indicator
#                              Indicates this is a telemetry packet
#                     (1 bit): Secondary flag
#                              Indicate presence of Secondary header
#                   (11 bits): ApID
#                              SPEXone ApID [0x320 - 0x351] or 2047
#
#  - Packet Sequence (2 bits): Grouping flag
#                              00 continuation packet-data segment
#                              01 first packet-data segment
#                              10 last packet-data segment
#                              11 packet-data unsegmented
#                   (14 bits): Counter per ApID, rollover to 0 at 0x3FFF
#  - Packet length  (16 bits): size of packet data in bytes (always odd)
#                              (secondary header + User data) - 1
#  - Packet timestamp:         Secondary header
#                   (32 bits): seconds (1-1-1970 UTC)
#                   (16 bits): sub-seconds (1/2 ** 16)

HDR_DTYPE = np.dtype(
    [
        ("type", ">u2"),
        ("sequence", ">u2"),
        ("length", ">u2"),
        ("tai_sec", ">u4"),
        ("sub_sec", ">u2"),
    ]
)

TIME_DTYPE = np.dtype([("tai_sec", ">u4"), ("sub_sec", ">u2")])

SCIHK_DTYPE = tmtc_dtype(0x350)

# - Error messages ------------------------
MSG_SKIP_FRAME = "[WARNING]: rejected a frame because it's incomplete"
MSG_INVALID_APID = (
    "[WARNING]: found one or more telemetry packages with an invalid APID"
)
MSG_CORRUPT_APID = "corrupted segments - detected APID 1 after <> 2"
MSG_CORRUPT_FRAME = "corrupted segments - previous frame not closed"


# - functions -----------------------------
def img_sec_of_day(
    img_sec: np.ndarray, img_subsec: np.ndarray, img_hk: np.ndarray
) -> tuple[datetime, float | Any]:
    """Convert Image CCSDS timestamp to seconds after midnight.

    Parameters
    ----------
    img_sec : numpy array (dtype='u4')
        Seconds since 1970-01-01 (or 1958-01-01)
    img_subsec : numpy array (dtype='u2')
        Sub-seconds as (1 / 2**16) seconds
    img_hk :  numpy array
        DemHK telemetry packages

    Returns
    -------
    tuple
        reference day: datetime, sec_of_day: numpy.ndarray

    """
    # determine for the first timestamp the offset with last midnight [seconds]
    epoch = datetime(1970, 1, 1, tzinfo=UTC)
    tstamp0 = epoch + timedelta(seconds=int(img_sec[0]))
    ref_day = datetime(
        year=tstamp0.year, month=tstamp0.month, day=tstamp0.day, tzinfo=UTC
    )
    # seconds since midnight
    offs_sec = (ref_day - epoch).total_seconds()

    # Determine offset wrt start-of-integration (IMRO + 1)
    # Where the default is defined as IMRO:
    #  [full-frame] COADDD + 2  (no typo, this is valid for the later MPS's)
    #  [binned] 2 * COADD + 1   (always valid)
    offs_msec = 0
    if img_hk["ICUSWVER"][0] > 0x123:
        imro = np.empty(img_hk.size, dtype=float)
        _mm = img_hk["IMRLEN"] == FULLFRAME_BYTES
        imro[_mm] = img_hk["REG_NCOADDFRAMES"][_mm] + 2
        imro[~_mm] = 2 * img_hk["REG_NCOADDFRAMES"][~_mm] + 1
        offs_msec = img_hk["FTI"] * (imro + 1) / 10

    # return seconds since midnight
    return ref_day, img_sec - offs_sec + img_subsec / 65536 - offs_msec / 1000


def hk_sec_of_day(
    ccsds_sec: np.ndarray, ccsds_subsec: np.ndarray, ref_day: datetime | None = None
) -> np.ndarray:
    """Convert CCSDS timestamp to seconds after midnight.

    Parameters
    ----------
    ccsds_sec : numpy array (dtype='u4')
        Seconds since 1970-01-01 (or 1958-01-01)
    ccsds_subsec : numpy array (dtype='u2')
        Sub-seconds as (1 / 2**16) seconds
    ref_day : datetime.datetime, optional
        `date` object represent a date (year, month, day)

    Returns
    -------
    numpy.ndarray with sec_of_day

    """
    # determine for the first timestamp the offset with last midnight [seconds]
    epoch = datetime(1970, 1, 1, tzinfo=UTC)
    if ref_day is None:
        tstamp0 = epoch + timedelta(seconds=int(ccsds_sec[0]))
        ref_day = datetime(
            year=tstamp0.year, month=tstamp0.month, day=tstamp0.day, tzinfo=UTC
        )
    offs_sec = (ref_day - epoch).total_seconds()

    # return seconds since midnight
    return ccsds_sec - offs_sec + ccsds_subsec / 65536


# - class CCSDSio -------------------------
class CCSDSio:
    """Read SPEXone telemetry packets.

    Parameters
    ----------
    file_list: iterator to strings
        list of file-names, where each file contains parts of a measurement

    Notes
    -----
    The formats of the PACE telemetry packets are following the standards:
    CCSDS-131.0-B-3, CCSDS-132.0-B-2 and CCSDS-133.0-B-1.

    This module is currenty restriced to telementry packets with APID:
    0x350 (Science), 0x320 (NomHK) and 0x322 (DemHK).

    A telemtry packet consist of a PRIMARY HEADER, SECONDARY HEADER
    (consist of a timestamp) and USER DATA with the actual telemetry
    packet data.

    Doc: TMTC handbook (SPX1-TN-005), issue 12, 15-May-2020

    The files with science and telemetry data needs to be in chronological
    order. However, you may mix science and housekeeping data as long as
    science data are chronological and housekeeping data are chronological.

    Examples
    --------
    >>> packets = ()
    >>> with CCSDSio(['file1', 'file2', 'file3']) as ccsds:
    >>>     while True:
    >>>         # read one telemetry packet at a time
    >>>         packet = ccsds.read_packet()
    >>>         if packet is None:
    >>>             # now we have read all files
    >>>             break
    >>>
    >>>         packets += (packet[0],)
    >>>
    >>>     # combine segmented Science packages
    >>>     science_tm = ccsds.science_tm(packets)
    >>>     # now you may want to collect the engineering packages

    """

    def __init__(self: CCSDSio, file_list: list) -> None:
        """Initialize access to a SPEXone Level-0 product (CCSDS format)."""
        # initialize class attributes
        self.file_list = iter(file_list)
        self.found_invalid_apid = False
        self.__hdr = None
        self.fp = None

        if file_list:
            self.open_next_file()

    def __iter__(self: CCSDSio) -> None:
        """Allow iteration."""
        for attr in sorted(self.__dict__):
            if not attr.startswith("__"):
                yield attr

    def __enter__(self: CCSDSio) -> CCSDSio:
        """Initiate the context manager."""
        return self

    def __exit__(self: CCSDSio, *args: str) -> bool:
        """Exit the context manager."""
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self: CCSDSio) -> None:
        """Close resources."""
        if self.fp is not None:
            if self.found_invalid_apid:
                print(MSG_INVALID_APID)
            self.found_invalid_apid = False
            self.fp.close()

    # ---------- define some class properties ----------
    @property
    def version_no(self: CCSDSio) -> int | None:
        """Returns CCSDS version number."""
        if self.__hdr is None:
            return None

        return (self.__hdr["type"] >> 13) & 0x7

    @property
    def type_indicator(self: CCSDSio) -> int | None:
        """Returns type of telemetry packet."""
        if self.__hdr is None:
            return None

        return (self.__hdr["type"] >> 12) & 0x1

    @property
    def secnd_hdr_flag(self: CCSDSio) -> bool | None:
        """Returns flag indicating presence of a secondary header."""
        if self.__hdr is None:
            return None

        return (self.__hdr["type"] >> 11) & 0x1

    @property
    def ap_id(self: CCSDSio) -> int | None:
        """Returns SPEXone ApID."""
        if self.__hdr is None:
            return None

        return self.__hdr["type"] & 0x7FF

    @property
    def grouping_flag(self: CCSDSio) -> int | None:
        """Returns grouping flag.

        The meaning of the grouping flag values are::

          00 continuation packet-data segment
          01 first packet-data segment
          10 last packet-data segment
          11 packet-data unsegmented

        """
        if self.__hdr is None:
            return None

        return (self.__hdr["sequence"] >> 14) & 0x3

    @property
    def sequence_count(self: CCSDSio) -> int | None:
        """Returns sequence counter, rollover to zero at 0x3FFF."""
        if self.__hdr is None:
            return None

        return self.__hdr["sequence"] & 0x3FFF

    @property
    def packet_length(self: CCSDSio) -> int | None:
        """Returns size of packet data in bytes.

        Value equals secondary header + user data (always odd)
        """
        if self.__hdr is None:
            return None

        return self.__hdr["length"]

    # ---------- define empty telemetry packet ----------
    def open_next_file(self: CCSDSio) -> None:
        """Open next file from file_list."""
        flname = next(self.file_list)
        if not Path(flname).is_file():
            raise FileNotFoundError(f"{flname} does not exist")

        self.close()
        # pylint: disable=consider-using-with
        self.fp = open(flname, "rb")  # noqa: SIM115

    @staticmethod
    def fix_dem_hk24(dem_hk: np.ndarray) -> np.ndarray:
        """Correct 32-bit values in the DemHk.

        Which originate from 24-bit values of the detector register parameters.

        Parameters
        ----------
        dem_hk : numpy.ndarray
           SPEXone DEM housekeeping packages

        Returns
        -------
        numpy.ndarray
           SPEXone DEM housekeeping packages

        """
        for key in [
            "DET_EXPTIME",
            "DET_EXPSTEP",
            "DET_KP1",
            "DET_KP2",
            "DET_EXPTIME2",
            "DET_EXPSTEP2",
        ]:
            dem_hk[key] = dem_hk[key] >> 8

        return dem_hk

    @staticmethod
    def fix_sci_hk24(sci_hk: np.ndarray) -> np.ndarray:
        """Correct 32-bit values in the Science HK.

        Which originate from 24-bit values in the detector register parameters.
        In addition::

         - copy the first 4 bytes of DET_CHENA to DET_ILVDS
         - parameter 'REG_BINNING_TABLE_START' was writen in little-endian

        Parameters
        ----------
        sci_hk : numpy.ndarray
           SPEXone Science telemetry packages

        Returns
        -------
        numpy.ndarray
           SPEXone Science telemetry packages

        """
        if np.all(sci_hk["ICUSWVER"] < 0x129):
            key = "REG_BINNING_TABLE_START"
            sci_hk[key] = np.ndarray(
                shape=sci_hk.shape, dtype="<u4", buffer=sci_hk[key]
            )

        sci_hk["DET_ILVDS"] = sci_hk["DET_CHENA"] & 0xF

        for key in [
            "TS1_DEM_N_T",
            "TS2_HOUSING_N_T",
            "TS3_RADIATOR_N_T",
            "TS4_DEM_R_T",
            "TS5_HOUSING_R_T",
            "TS6_RADIATOR_R_T",
            "LED1_ANODE_V",
            "LED1_CATH_V",
            "LED1_I",
            "LED2_ANODE_V",
            "LED2_CATH_V",
            "LED2_I",
            "ADC1_VCC",
            "ADC1_REF",
            "ADC1_T",
            "ADC2_VCC",
            "ADC2_REF",
            "ADC2_T",
            "DET_EXPTIME",
            "DET_EXPSTEP",
            "DET_KP1",
            "DET_KP2",
            "DET_EXPTIME2",
            "DET_EXPSTEP2",
            "DET_CHENA",
        ]:
            sci_hk[key] = sci_hk[key] >> 8

        return sci_hk

    def __rd_science(self: CCSDSio, hdr: np.ndarray) -> np.ndarray:
        """Read Science telemetry packet.

        Parameters
        ----------
        hdr :  numpy.ndarray
           CCSDS header information

        Returns
        -------
        numpy.ndarray
           SPEXone Science telemetry packages

        """
        num_bytes = self.packet_length - TIME_DTYPE.itemsize + 1
        packet = np.empty(
            1,
            dtype=np.dtype(
                [
                    ("packet_header", HDR_DTYPE),
                    ("science_hk", SCIHK_DTYPE),
                    ("icu_time", TIME_DTYPE),
                    ("image_data", "O"),
                ]
            ),
        )
        packet["packet_header"] = hdr

        # first segment or unsegmented data packet provides Science_HK
        if self.grouping_flag in (1, 3):
            packet["science_hk"] = self.fix_sci_hk24(
                np.fromfile(self.fp, count=1, dtype=SCIHK_DTYPE)
            )
            num_bytes -= SCIHK_DTYPE.itemsize
            packet["icu_time"] = np.fromfile(self.fp, count=1, dtype=TIME_DTYPE)
            num_bytes -= TIME_DTYPE.itemsize

        # read detector image data
        packet["image_data"][0] = np.fromfile(
            self.fp, dtype=">u2", count=num_bytes // 2
        )
        return packet

    def __rd_nomhk(self: CCSDSio, hdr: np.ndarray) -> np.ndarray:
        """Read NomHK telemetry packet.

        Parameters
        ----------
        hdr :  numpy.ndarray
           CCSDS header information

        Returns
        -------
        numpy.ndarray
           SPEXone Nominal housekeeping packages

        """
        packet = np.empty(
            1,
            dtype=np.dtype(
                [("packet_header", HDR_DTYPE), ("nominal_hk", tmtc_dtype(0x320))]
            ),
        )
        packet["packet_header"] = hdr
        packet["nominal_hk"] = np.fromfile(self.fp, count=1, dtype=tmtc_dtype(0x320))
        return packet

    def __rd_demhk(self: CCSDSio, hdr: np.ndarray) -> np.ndarray:
        """Read DemHK telemetry packet.

        Parameters
        ----------
        hdr :  numpy.ndarray
           CCSDS header information

        Returns
        -------
        numpy.ndarray
           SPEXone detector housekeeping packages

        """
        packet = np.empty(
            1,
            dtype=np.dtype(
                [("packet_header", HDR_DTYPE), ("detector_hk", tmtc_dtype(0x322))]
            ),
        )
        packet["packet_header"] = hdr
        packet["detector_hk"] = self.fix_dem_hk24(
            np.fromfile(self.fp, count=1, dtype=tmtc_dtype(0x322))
        )
        return packet

    def __rd_tc_accept(self: CCSDSio, _: np.ndarray) -> np.ndarray:
        """Read/dump TcAccept packet."""
        self.fp.seek(-1 * HDR_DTYPE.itemsize, 1)
        packet = np.fromfile(
            self.fp,
            count=1,
            dtype=np.dtype(
                [
                    ("packet_header", HDR_DTYPE),
                    ("TcPacketId", ">u2"),
                    ("TcSeqControl", ">u2"),
                ]
            ),
        )
        print("[TcAccept]: ", self, packet["TcPacketId"][0], packet["TcSeqControl"][0])
        return packet

    def __rd_tc_execute(self: CCSDSio, _: np.ndarray) -> np.ndarray:
        """Read/dump TcExecute packet."""
        self.fp.seek(-1 * HDR_DTYPE.itemsize, 1)
        packet = np.fromfile(
            self.fp,
            count=1,
            dtype=np.dtype(
                [
                    ("packet_header", HDR_DTYPE),
                    ("TcPacketId", ">u2"),
                    ("TcSeqControl", ">u2"),
                ]
            ),
        )
        print("[TcExecute]:", self, packet["TcPacketId"][0], packet["TcSeqControl"][0])
        return packet

    def __rd_tc_fail(self: CCSDSio, _: np.ndarray) -> np.ndarray:
        """Read/dump TcFail packet."""
        self.fp.seek(-1 * HDR_DTYPE.itemsize, 1)
        packet = np.fromfile(
            self.fp,
            count=1,
            dtype=np.dtype(
                [
                    ("packet_header", HDR_DTYPE),
                    ("TcPacketId", ">u2"),
                    ("TcSeqControl", ">u2"),
                    ("TcFailCode", ">u2"),
                    ("FailParameter1", ">u2"),
                    ("FailParameter2", ">u2"),
                ]
            ),
        )
        print(
            "[TcFail]:   ",
            self,
            packet["TcPacketId"][0],
            packet["TcSeqControl"][0],
            bin(packet["TcFailCode"][0]),
            packet["FailParameter1"][0],
            packet["FailParameter2"][0],
        )
        return packet

    def __rd_tc_reject(self: CCSDSio, _: np.ndarray) -> np.ndarray:
        """Read/dump TcReject packet."""
        self.fp.seek(-1 * HDR_DTYPE.itemsize, 1)
        packet = np.fromfile(
            self.fp,
            count=1,
            dtype=np.dtype(
                [
                    ("packet_header", HDR_DTYPE),
                    ("TcPacketId", ">u2"),
                    ("TcSeqControl", ">u2"),
                    ("TcRejectCode", ">u2"),
                    ("RejectParameter1", ">u2"),
                    ("RejectParameter2", ">u2"),
                ]
            ),
        )
        print(
            "[TcReject]: ",
            self,
            packet["TcPacketId"][0],
            packet["TcSeqControl"][0],
            bin(packet["TcRejectCode"][0]),
            packet["RejectParameter1"][0],
            packet["RejectParameter2"][0],
        )
        return packet

    def __rd_other(self: CCSDSio, hdr: np.ndarray) -> np.ndarray | None:
        """Read other telemetry packet."""
        num_bytes = self.packet_length - TIME_DTYPE.itemsize + 1
        if not 0x320 <= self.ap_id <= 0x350:
            self.found_invalid_apid = True
            self.fp.seek(num_bytes, 1)
            return None

        packet = np.empty(
            1,
            dtype=np.dtype(
                [("packet_header", HDR_DTYPE), ("raw_data", "u1", num_bytes)]
            ),
        )
        packet["packet_header"] = hdr
        packet["raw_data"] = np.fromfile(self.fp, count=num_bytes, dtype="u1")
        return packet

    def read_packet(self: CCSDSio) -> np.ndarray | None:
        """Read next telemetry packet.

        Returns
        -------
        numpy.ndarray
           CCSDS packet data

        """
        if self.fp is None:
            return None

        # read primary/secondary header
        hdr = np.fromfile(self.fp, count=1, dtype=HDR_DTYPE)
        if hdr.size == 0:
            try:
                self.open_next_file()
            except StopIteration:
                return None
            except FileNotFoundError:
                return None

            hdr = np.fromfile(self.fp, count=1, dtype=HDR_DTYPE)
            if hdr.size == 0:
                return None

        # save packet header as class attribute
        self.__hdr = hdr[0]

        # read telemetry packet
        rd_func = {
            0x350: self.__rd_science,
            0x320: self.__rd_nomhk,
            0x322: self.__rd_demhk,
            0x331: self.__rd_tc_accept,
            0x332: self.__rd_tc_reject,
            0x333: self.__rd_tc_execute,
            0x334: self.__rd_tc_fail,
        }.get(self.ap_id, self.__rd_other)
        return rd_func(hdr)

    @staticmethod
    def select_tm(packets_in: tuple, ap_id: int) -> tuple:
        """Select telemetry packages on SPEXone ApID.

        Parameters
        ----------
        packets_in: tuple
           SPEXone telemetry packages
        ap_id: int
           SPEXone ApID

        Returns
        -------
        tuple
           selected telemetry packages

        """
        packets = ()
        for packet in packets_in:
            if "packet_header" not in packet.dtype.names:
                continue

            if (packet["packet_header"]["type"] & 0x7FF) == ap_id:
                packets += (packet,)

        return packets

    def science_tm(self: CCSDSio, packets_in: tuple) -> tuple:
        """Combine segmented Science telemetry packages.

        Parameters
        ----------
        packets_in: tuple
           Science or house-keeping telemetry packages

        Returns
        -------
        tuple
           unsegmented Science telemetry packages

        """
        # reject non-Science telemetry packages
        packets = self.select_tm(packets_in, 0x350)
        if not packets:
            return ()

        # check if grouping_flag of first segment equals 1
        #   else reject all segments with grouping_flag != 1
        self.__hdr = packets[0]["packet_header"]
        if self.grouping_flag != 1:
            ii = 0
            for packet in packets:
                self.__hdr = packet["packet_header"]
                if self.grouping_flag == 1:
                    break
                ii += 1

            print(f"[WARNING]: first frame incomplete - skipped {ii} segments")
            packets = packets[ii:]
            if not packets:
                return ()

        # check if grouping_flag of last segment equals 2
        #   else reject all segments after the last segment
        #   with grouping_flag == 2
        self.__hdr = packets[-1]["packet_header"]
        if self.grouping_flag != 2:
            ii = 0
            for packet in packets:
                self.__hdr = packet["packet_header"]
                if self.grouping_flag == 2:
                    break
                ii += 1

            print(f"[WARNING]: last frame incomplete - rejected {ii} segments")
            packets = packets[:-ii]
            if not packets:
                return ()

        res = ()
        offs = 0
        prev_grp_flag = 2
        for packet in packets:
            grouping_flag = (packet["packet_header"]["sequence"] >> 14) & 0x3
            # print(prev_grp_flag, grouping_flag, len(res), offs,
            #      packet['image_data'].size)
            # handle segmented data
            if grouping_flag == 1:  # first segment
                # group_flag of previous package should be 2
                if prev_grp_flag != 2:
                    if packet["image_data"].size in (3853, 7853):
                        print(MSG_SKIP_FRAME)
                        prev_grp_flag = 2
                        offs = 0
                    else:
                        raise RuntimeError(MSG_CORRUPT_APID)

                img_size = packet["science_hk"]["IMRLEN"] // 2
                rec_buff = np.empty(
                    1,
                    dtype=np.dtype(
                        [
                            ("packet_header", HDR_DTYPE),
                            ("science_hk", SCIHK_DTYPE),
                            ("icu_time", TIME_DTYPE),
                            ("image_data", "O"),
                        ]
                    ),
                )[0]
                img_buff = np.empty(img_size, dtype="u2")

                rec_buff["packet_header"] = packet["packet_header"]
                rec_buff["science_hk"] = packet["science_hk"]
                rec_buff["icu_time"] = packet["icu_time"]
                img_buff[offs : offs + packet["image_data"].size] = packet["image_data"]
                offs += packet["image_data"].size
            else:  # continuation or last segment
                # group_flag of previous package should be 0 or 1
                if prev_grp_flag == 2:
                    raise RuntimeError(MSG_CORRUPT_FRAME)

                img_buff[offs : offs + packet["image_data"].size] = packet["image_data"]
                offs += packet["image_data"].size
                if grouping_flag == 2:
                    if offs == img_size:
                        rec_buff["image_data"] = img_buff
                        res += (rec_buff,)
                    else:
                        print(MSG_SKIP_FRAME)
                    offs = 0

            # keep current group flag for next read
            prev_grp_flag = grouping_flag

        return res



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2022-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# Author: Sean Bailey - DAAC Manager, OB.DAAC, NASA/GSFC Code 619
# License:  BSD-3-Clause
"""Python script to determine the number of leap seconds for given timestamp.

The source for the latest version of tai-utc.dat is the US Naval Observatory:

   https://maia.usno.navy.mil/ser7/tai-utc.dat
"""

from __future__ import annotations

__all__ = ["get_leap_seconds"]

import datetime as dt
import logging
from importlib.resources import files
from os import environ
from pathlib import Path

import julian

# - global parameters -----------------------
module_logger = logging.getLogger("pyspex.lib.leap_sec")


def get_leap_seconds(taitime: float, epochyear: int = 1958) -> float:
    """Return the number of elapsed leap seconds given a TAI time in seconds.

    Requires the file tai-utc.dat.
    """
    # determine location of the file 'tai-utc.dat'
    ocvarroot = environ.get("OCVARROOT")
    if ocvarroot is None:
        taiutc = files("pyspex.data").joinpath("tai-utc.dat")
    else:
        taiutc = Path(ocvarroot) / "common" / "tai-utc.dat"

    epochsecs = dt.datetime(epochyear, 1, 1, tzinfo=dt.UTC).timestamp()
    taidt = dt.datetime.fromtimestamp(taitime + epochsecs, dt.UTC)
    leapsec: float = 0
    with taiutc.open("r", encoding="ascii") as fp:
        for line in fp:
            rec = line.rstrip().split(None, 7)
            if julian.from_jd(float(rec[4])).replace(tzinfo=dt.UTC) < taidt:
                leapsec = float(rec[6])
    module_logger.debug("leap seconds: %f", leapsec)

    return leapsec



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Collect EGSE/OGSE data and/or add this to a SPEXone Level-1A product."""

from __future__ import annotations

__all__ = ["LV1gse"]

from typing import TYPE_CHECKING, Any

import numpy as np

# pylint: disable=no-name-in-module
from netCDF4 import Dataset

if TYPE_CHECKING:
    import xarray as xr


# --------------------------------------------------
class LV1gse:
    """Adds EGSE/OGSE data to a SPEXone Level-1A product.

    Parameters
    ----------
    l1a_file: str
        Name of the Level-1A product

    """

    def __init__(self: LV1gse, l1a_file: str) -> None:
        """Initialize netCDF4 group 'gse_data' in a SPEXone Level-1 product."""
        self.fid = Dataset(l1a_file, "r+")

        if self.fid.groups.get("gse_data"):
            return

        # investigate filename
        parts = l1a_file.split("_")
        if len(parts) > 2 and parts[0] == "SPX1":
            parts = parts[2:]
        msmt_fields = parts[0].split("-")
        background = "BKG" in msmt_fields
        act_angle = [float(x.replace("act", "")) for x in parts if x.startswith("act")]
        alt_angle = [float(x.replace("alt", "")) for x in parts if x.startswith("alt")]
        pol_angle = [
            float(x.replace("pol", ""))
            for x in parts
            if x.startswith("pol") and x != "polcal"
        ]
        gp1_angle = [
            float(x.replace("glass", "")) for x in parts if x.startswith("glass")
        ]
        gp1_offs = 5.634375
        # gp2_offs = 5.09625

        # determine viewport: default 0, when all viewports are illuminated
        if alt_angle:
            vp_angle = np.array([-50.0, -20, 0, 20, 50])
            vp_diff = np.abs(vp_angle - alt_angle[0])
            viewport = 0 if vp_diff.min() > 6 else 2 ** np.argmin(vp_diff)
        else:
            viewport = 0

        gid = self.fid.createGroup("/gse_data")
        dset = gid.createVariable("viewport", "u1")
        dset.long_name = "viewport status"
        dset.standard_name = "status_flag"
        dset.valid_range = np.array([0, 16], dtype="u1")
        dset.flag_values = np.array([0, 1, 2, 4, 8, 16], dtype="u1")
        dset.flag_meanings = "ALL -50deg -20deg 0deg +20deg +50deg"
        dset[:] = viewport

        # gid.FOV_begin = np.nan
        # gid.FOV_end = np.nan
        gid.ACT_rotationAngle = np.nan if not act_angle else act_angle[0]
        gid.ALT_rotationAngle = np.nan if not alt_angle else alt_angle[0]
        # gid.ACT_illumination = np.nan
        # gid.ALT_illumination = np.nan
        gid.AoLP = 0.0 if not pol_angle else pol_angle[0]
        if not background and msmt_fields[0] in ("POLARIZED", "POLARIMETRIC"):
            gid.DoLP = 1.0
            if gp1_angle:
                gid.GP1_angle = gp1_angle[0] + gp1_offs
        else:
            gid.DoLP = 0.0

    def __iter__(self: LV1gse) -> None:
        """Allow iteration."""
        for attr in sorted(self.__dict__):
            if not attr.startswith("__"):
                yield attr

    def __enter__(self: LV1gse) -> LV1gse:
        """Initiate the context manager."""
        return self

    def __exit__(self: LV1gse, *args: str) -> bool:
        """Exit the context manager."""
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self: LV1gse) -> None:
        """Close all resources (currently a placeholder function)."""
        if self.fid is None:
            return

        self.fid.close()
        self.fid = None

    def check_egse(self: LV1gse, egse_data: np.ndarray) -> None:
        """Check consistency of OGSE/EGSE information during measurement."""
        for key, fmt in egse_data.dtype.fields.items():
            if fmt[0] == np.uint8:
                res_sanity = np.all(egse_data[key] == egse_data[key][0])
                if not res_sanity:
                    print(f"[WARNING]: {key}={egse_data[key]}")

        act_angle = self.fid["/gse_data"].ACT_rotationAngle
        if np.isfinite(act_angle) and not np.allclose(
            egse_data["ACT_ANGLE"], act_angle, 1e-2
        ):
            print(f'[WARNING]: ACT_ANGLE={egse_data["ACT_ANGLE"]}')

        alt_angle = self.fid["/gse_data"].ALT_rotationAngle
        if np.isfinite(alt_angle) and not np.allclose(
            egse_data["ALT_ANGLE"], alt_angle, 1e-2
        ):
            print(f'[WARNING]: ALT_ANGLE={egse_data["ALT_ANGLE"]}')

    def set_attr(self: LV1gse, name: str, value: Any) -> None:  # noqa: ANN401
        """Add attribute to group 'gse_data'.

        Parameters
        ----------
        name :  str
           name of the attribute
        value :  Any
           value of the attribute

        """
        self.fid["/gse_data"].setncattr(name, value)

    def write_attr_act(
        self: LV1gse, angle: float, illumination: float | None = None
    ) -> None:
        """Add act rotation angle as a group attribute.

        Parameters
        ----------
        angle :  float
           ACT angle
        illumination :  float, optional
           Illumination level

        """
        self.fid["/gse_data"].ACT_rotationAngle = angle
        if illumination is not None:
            self.fid["/gse_data"].ACT_illumination = illumination

    def write_attr_alt(
        self: LV1gse, angle: float, illumination: float | None = None
    ) -> None:
        """Add altitude rotation angle as a group attribute.

        Parameters
        ----------
        angle :  float
           ALT angle
        illumination :  float
           Illumination level

        """
        self.fid["/gse_data"].ALT_rotationAngle = angle
        if illumination is not None:
            self.fid["/gse_data"].ALT_illumination = illumination

    def write_attr_polarization(self: LV1gse, aolp: float, dolp: float) -> None:
        """Add polarization parameters AoLP & DoLP as group attributes.

        Parameters
        ----------
        aolp :  float
           Angle of linear polarization
        dolp :  float
           Degree of linear polarization

        """
        if aolp is not None:
            self.fid["/gse_data"].AoLP = aolp
        if dolp is not None:
            self.fid["/gse_data"].DoLP = dolp

    def write_egse(
        self: LV1gse, egse_time: np.ndarray, egse_data: np.ndarray, egse_attrs: dict
    ) -> None:
        """Add EGSE parameters.

        Parameters
        ----------
        egse_time :  ndarray
           EGSE timestamps
        egse_data :  ndarray
           EGSE data
        egse_attrs :  dict
           EGSE attributes

        """
        # perform sanity check on EGSE parameters
        self.check_egse(egse_data)

        gid = self.fid["/gse_data"]
        _ = gid.createDimension("time", len(egse_data))
        dset = gid.createVariable("time", "f8", ("time",))
        dset[:] = egse_time

        egse_t = gid.createCompoundType(egse_data.dtype, "egse_dtype")
        dset = gid.createVariable("egse", egse_t, ("time",))
        dset.long_name = egse_attrs["long_name"]
        dset.fields = egse_attrs["fields"]
        dset.units = egse_attrs["units"]
        dset.comment = egse_attrs["comment"]
        dset[:] = egse_data

    def write_viewport(self: LV1gse, viewport: int) -> None:
        """Add/update which viewports are illuminated.

        Parameters
        ----------
        viewport :  int
           Viewport as an intergier between 0 and 5

        """
        self.fid["/gse_data/viewport"][:] = viewport

    def write_reference_signal(
        self: LV1gse, signal: np.ndarray, stdev: np.ndarray
    ) -> None:
        """Write reference detector signal and variance.

        Parameters
        ----------
        signal :  float
           Mean of signal level measured during the measurement
        stdev :  float
           standard deviation of signal level measured during the measurement

        Notes
        -----
        Used for non-linearity measurements.

        """
        gid = self.fid["/gse_data"]
        gid.Illumination_level = 5e9 / 1.602176634 * signal

        dset = gid.createVariable("reference_signal", "f8", ())
        dset.long_name = "mean of reference-diode signal"
        dset.comment = "t_sat = min(2.28e-9 / S_reference, 30)"
        dset.units = "A"
        dset[:] = signal

        dset = gid.createVariable("reference_signal_std", "f8", ())
        dset.long_name = "standard deviation of reference-diode signal"
        dset.units = "A"
        dset[:] = stdev

    def write_reference_diode(
        self: LV1gse, ref_time: np.ndarray, ref_data: np.ndarray, ref_attrs: dict
    ) -> None:
        """Add data measured by the reference diode during the measurement.

        Parameters
        ----------
        ref_time :  ndarray
           reference diode timestamps
        ref_data :  ndarray
           reference diode data
        ref_attrs :  dict
           reference diode attributes
           fid['/ReferenceDiode/ref_diode'].__dict__

        """
        gid = self.fid.createGroup("/gse_data/ReferenceDiode")
        _ = gid.createDimension("time", len(ref_data))
        dset = gid.createVariable("time", "f8", ("time",))
        dset.units = "seconds since 1970-01-01 00:00:00"
        dset.comment = "Generated on SRON clean-room tablet"
        dset[:] = ref_time

        ref_t = gid.createCompoundType(ref_data.dtype, "ref_dtype")
        dset = gid.createVariable("ref_diode", ref_t, ("time",))
        dset.setncatts(ref_attrs)
        dset[:] = ref_data

    def write_wavelength_monitor(self: LV1gse, xds_wav_mon: xr.Dataset) -> None:
        """Add wavelength monitoring data of the Avantas fibre-spectrometer.

        Parameters
        ----------
        xds_wav_mon :  xarray::Dataset
           Contains xarray::DataArrays 'signal', 'wavelength', 'wav_time'

        """
        xds_wav_mon.to_netcdf(
            self.fid.filepath(), mode="a", group="/gse_data/WaveMonitor"
        )

        # gid = self.fid.createGroup('/gse_data/WaveMonitor')
        # _ = gid.createDimension('time', len(wav_time))
        # dset = gid.createVariable('time', 'f8', ('time',))
        # dset.units = 'seconds since 1970-01-01 00:00:00'
        # dset.comment = 'Generated on SRON clean-room tablet'
        # dset[:] = wav_time
        # _ = gid.createDimension('wavelength', len(wav_wv))
        # dset = gid.createVariable('wavelength', 'f4', ('wavelength',))
        # dset.longname = 'wavelength grid'
        # dset.comment = 'wavelength annotation of fibre spectrometer'
        # dset[:] = wav_wv
        # dset = gid.createVariable('t_intg', 'i2', ('time',))
        # dset.long_name = 'Integration time'
        # dset.units = 'ms'
        # dset[:] = wav_intg
        # dset = gid.createVariable('n_avg', 'i2', ('time',))
        # dset.long_name = 'Averaging number'
        # dset[:] = wav_avg_num
        # dset = gid.createVariable('wav_mon', 'f4', ('time', 'wavelength'))
        # dset.long_name = 'wavelength-monitor spectra'
        # dset.comment = 'Avantes fibre spectrometer'
        # dset[:] = wav_signal
